
package de.csw.ludum.dare.ld23.util;

import java.awt.Rectangle;

import de.csw.ludum.dare.ld23.graphic.Bitmap;

public class BitmapLine {

  private final static int PRECISION_SHIFT = 8;

  /**
   * Draws a colored line by connecting two points using a DDA algorithm
   * (Digital Differential Analyzer)
   */
  public static void drawLineDDA(Bitmap bmp, int x1, int y1, int x2, int y2, CheapColor color) {
    drawLineDDA(bmp, x1, y1, x2, y2, color.colorIntValue);
  }

  /**
   * Draws a colored line by connecting two points using a DDA algorithm
   * (Digital Differential Analyzer)
   */
  public static void drawLineDDA(Bitmap bmp, int x1, int y1, int x2, int y2, int color) {
    // Use refs for faster access (really important!) speeds up a lot!
    int w = bmp.width;
    int h = bmp.height;

    // Distance start and end point
    int dx = x2 - x1;
    int dy = y2 - y1;

    // Determine slope (absoulte value)
    int len = dy >= 0 ? dy : -dy;
    int lenx = dx >= 0 ? dx : -dx;
    if (lenx > len) {
      len = lenx;
    }

    // Prevent divison by zero
    if (len != 0) {
      // Init steps and start
      float incx = dx / (float) len;
      float incy = dy / (float) len;
      float x = x1;
      float y = y1;

      // Walk the line!
      for (int i = 0; i < len; i++) {
        if (y < h && y >= 0 && x < w && x >= 0) {
          bmp.pixels[(int) y * w + (int) x] = color;
        }
        x += incx;
        y += incy;
      }
    }
  }

  public static void drawLine(Bitmap bitmap, int x1, int y1, int x2, int y2, CheapColor color) {
    drawLine(bitmap, 10, 10, x1, y1, x2, y2, color.colorIntValue);
  }

  public static void drawLine(Bitmap bitmap, int pixelWidth, int pixelHeight, int x1, int y1, int x2, int y2, int color) {
    // Distance start and end point
    int dx = x2 - x1;
    int dy = y2 - y1;

    // Determine slope (absoulte value)
    int lenX, lenY;
    if (dy >= 0) {
      lenY = dy;
    } else {
      lenY = -dy;
    }

    if (dx >= 0) {
      lenX = dx;
    } else {
      lenX = -dx;
    }

    if (lenX > lenY) { // x increases by +/- 1
      if (dx < 0) {
        int t = x1;
        x1 = x2;
        x2 = t;
        t = y1;
        y1 = y2;
        y2 = t;
      }

      // Init steps and start
      int incy = (dy << PRECISION_SHIFT) / dx;

      int y1s = y1 << PRECISION_SHIFT;
      int y2s = y2 << PRECISION_SHIFT;
      int hs = pixelHeight << PRECISION_SHIFT;

      if (y1 < y2) {
        if (y1 >= pixelHeight || y2 < 0) {
          return;
        }
        if (y1s < 0) {
          if (incy == 0) {
            return;
          }
          int oldy1s = y1s;
          // Find lowest y1s that is greater or equal than 0.
          y1s = incy - 1 + ((y1s + 1) % incy);
          x1 += (y1s - oldy1s) / incy;
        }
        if (y2s >= hs) {
          if (incy != 0) {
            // Find highest y2s that is less or equal than ws - 1.
            // y2s = y1s + n * incy. Find n.
            y2s = hs - 1 - (hs - 1 - y1s) % incy;
            x2 = x1 + (y2s - y1s) / incy;
          }
        }
      } else {
        if (y2 >= pixelHeight || y1 < 0) {
          return;
        }
        if (y1s >= hs) {
          if (incy == 0) {
            return;
          }
          int oldy1s = y1s;
          // Find highest y1s that is less or equal than ws - 1.
          // y1s = oldy1s + n * incy. Find n.
          y1s = hs - 1 + (incy - (hs - 1 - oldy1s) % incy);
          x1 += (y1s - oldy1s) / incy;
        }
        if (y2s < 0) {
          if (incy != 0) {
            // Find lowest y2s that is greater or equal than 0.
            // y2s = y1s + n * incy. Find n.
            y2s = y1s % incy;
            x2 = x1 + (y2s - y1s) / incy;
          }
        }
      }

      if (x1 < 0) {
        y1s -= incy * x1;
        x1 = 0;
      }
      if (x2 >= pixelWidth) {
        x2 = pixelWidth - 1;
      }

      int ys = y1s;

      // Walk the line!
      int y = ys >>> PRECISION_SHIFT;
      int previousY = y;
      int index = x1 + y * pixelWidth;
      int k = incy < 0 ? 1 - pixelWidth : 1 + pixelWidth;
      for (int x = x1; x <= x2; ++x) {
        bitmap.pixels[index] = color;
        ys += incy;
        y = ys >>> PRECISION_SHIFT;
        if (y != previousY) {
          previousY = y;
          index += k;
        } else {
          ++index;
        }
      }
    } else {
      // Prevent divison by zero
      if (lenY == 0) {
        return;
      }
      if (dy < 0) {
        int t = x1;
        x1 = x2;
        x2 = t;
        t = y1;
        y1 = y2;
        y2 = t;
      }

      // Init steps and start
      int x1s = x1 << PRECISION_SHIFT;
      int x2s = x2 << PRECISION_SHIFT;
      int ws = pixelWidth << PRECISION_SHIFT;

      int incx = (dx << PRECISION_SHIFT) / dy;

      if (x1 < x2) {
        if (x1 >= pixelWidth || x2 < 0) {
          return;
        }
        if (x1s < 0) {
          if (incx == 0) {
            return;
          }
          int oldx1s = x1s;
          // Find lowest x1s that is greater or equal than 0.
          x1s = incx - 1 + ((x1s + 1) % incx);
          y1 += (x1s - oldx1s) / incx;
        }
        if (x2s >= ws) {
          if (incx != 0) {
            // Find highest x2s that is less or equal than ws - 1.
            // x2s = x1s + n * incx. Find n.
            x2s = ws - 1 - (ws - 1 - x1s) % incx;
            y2 = y1 + (x2s - x1s) / incx;
          }
        }
      } else {
        if (x2 >= pixelWidth || x1 < 0) {
          return;
        }
        if (x1s >= ws) {
          if (incx == 0) {
            return;
          }
          int oldx1s = x1s;
          // Find highest x1s that is less or equal than ws - 1.
          // x1s = oldx1s + n * incx. Find n.
          x1s = ws - 1 + (incx - (ws - 1 - oldx1s) % incx);
          y1 += (x1s - oldx1s) / incx;
        }
        if (x2s < 0) {
          if (incx != 0) {
            // Find lowest x2s that is greater or equal than 0.
            // x2s = x1s + n * incx. Find n.
            x2s = x1s % incx;
            y2 = y1 + (x2s - x1s) / incx;
          }
        }
      }

      if (y1 < 0) {
        x1s -= incx * y1;
        y1 = 0;
      }
      if (y2 >= pixelHeight) {
        y2 = pixelHeight - 1;
      }

      int index = x1s + ((y1 * pixelWidth) << PRECISION_SHIFT);

      // Walk the line!
      int inc = (pixelWidth << PRECISION_SHIFT) + incx;
      for (int y = y1; y <= y2; ++y) {
        bitmap.pixels[index >>> PRECISION_SHIFT] = color;
        index += inc;
      }
    }
  }

  public static void drawRectangle(Bitmap bmp, Rectangle rect, CheapColor color) {
    drawRectangle(bmp, rect.x, rect.y, rect.x + rect.width, rect.y + rect.height, color);
  }

  private static void drawRectangle(Bitmap bmp, int x1, int y1, int x2, int y2, CheapColor color) {
    drawRectangle(bmp, x1, y1, x2, y2, color.colorIntValue);
  }

  public static void drawRectangle(Bitmap bmp, int x1, int y1, int x2, int y2, int color) {
    // Use refs for faster access (really important!) speeds up a lot!
    int w = bmp.width;
    int h = bmp.height;
    try {
      // Check boundaries
      if ((x1 < 0 && x2 < 0) || (y1 < 0 && y2 < 0) || (x1 >= w && x2 >= w) || (y1 >= h && y2 >= h)) {
        return;
      }

      // Clamp boundaries
      if (x1 < 0) {
        x1 = 0;
      }
      if (y1 < 0) {
        y1 = 0;
      }
      if (x2 < 0) {
        x2 = 0;
      }
      if (y2 < 0) {
        y2 = 0;
      }
      if (x1 > w) {
        x1 = w;
      }
      if (y1 > h) {
        y1 = h;
      }
      if (x2 > w) {
        x2 = w;
      }
      if (y2 > h) {
        y2 = h;
      }

      int startY = y1 * w;
      int endY = y2 * w;

      int offset2 = (endY - w) + x1;
      int endOffset = startY + x2;
      int startYPlusX1 = startY + x1;

      // top and bottom horizontal scanlines
      for (int x = startYPlusX1; x < endOffset; x++) {
        bmp.pixels[x] = color; // top horizontal line
        bmp.pixels[offset2] = color; // bottom horizontal line
        offset2++;
      }

      // offset2 == endY + x2

      // vertical scanlines
      endOffset = startYPlusX1 + w;
      offset2 -= w;

      for (int y = startY + x2 - 1 + w; y < offset2; y += w) {
        bmp.pixels[y] = color; // right vertical line
        bmp.pixels[endOffset] = color; // left vertical line
        endOffset += w;
      }

    } catch (Exception e) {
      e.printStackTrace();
    }
  }
}
